// Mother **************************************************************
//	George Marsaglia's The mother of all random number generators
//	producing uniformly distributed pseudo random 32 bit values with
//	period about 2^250.
//
//	Bob Wheeler 8/8/94
//	Re-arranged for the BeOS RNG Library by John Cooke 1998-04-21
//

#include "mother.h"
#include <string.h>

#define m16Long 65536L				/* 2^16 */
#define m16Mask 0xFFFF          /* mask for lower 16 bits */
#define m15Mask 0x7FFF			/* mask for lower 15 bits */
#define m31Mask 0x7FFFFFFF     /* mask for 31 bits */
#define m32Double  4294967295.0  /* 2^32-1 */
#define m32Drecip  (1.0/m32Double)  /* the reciprocal of m32Double*/

Mother::Mother (uint32 seed) { Seed(seed); }

void Mother::Seed (uint32 seed)
{
	unsigned short half;
	uint32 whole;
	unsigned short *p;
	int n;

/* Initialize the bufs with 9 random values the first time */
	half = seed&m16Mask;					/* The low 16 bits */
	whole = seed&m31Mask;					/* Only want 31 bits */

/* fudge if whole is zero */
	if(whole == 0)
		whole = 0x7fff0000;

	p = buffer;
	for (n=18;n--;) {
		whole = 30903*half+(whole>>16);		/* One line multiply-with-carry */
		*p++ = half = whole&m16Mask;
		if(n == 9)
			p = buffer+10;
	}

/* make carry 15 bits */
	buffer[0] &= m15Mask;
	buffer[10] &= m15Mask;
}

uint32 Mother::lRand()
{
	uint32 top, bot;

/* treat the buffer as if it were two buffers */
	unsigned short *topbuf = buffer+0;
	unsigned short *botbuf = buffer+10;

/* Move elements 1 to 8 to 2 to 9 */
	memmove(topbuf+2,topbuf+1,8*sizeof(short));
	memmove(botbuf+2,botbuf+1,8*sizeof(short));

/* Put the carry values in top+bot */
	top = topbuf[0];
	bot = botbuf[0];

/* Form the linear combinations */

	top += 1941*topbuf[2]+1860*topbuf[3]+1812*topbuf[4]+1776*topbuf[5]+
		1492*topbuf[6]+1215*topbuf[7]+1066*topbuf[8]+12013*topbuf[9];

	bot += 1111*botbuf[2]+2222*botbuf[3]+3333*botbuf[4]+4444*botbuf[5]+
		5555*botbuf[6]+6666*botbuf[7]+7777*botbuf[8]+9272*botbuf[9];

/* Save the high bits of top+bot as the new carry */
	topbuf[0] = top/m16Long;
	botbuf[0] = bot/m16Long;

/* Put the low bits of top+bot into buf[1] */
	topbuf[1] = m16Mask&top;
	botbuf[1] = m16Mask&bot;

/* Combine the two 16 bit random numbers into one 32 bit */
	return (((long)topbuf[1])<<16)+(long)botbuf[1];
}

double Mother::dRand()
{
/* Return a double value between 0.0 and <1.0 */
#ifdef USE_DIVIDE
	return lRand()/m32Double;
#else
	return lRand() * m32Drecip;	/* multiply is faster than divide */
#endif
}

// this is the Info about the Mother RNG
static RNGInfoItem mothitems[] =
{
	{ "Class", "Mother"},
	{ "Description", "The Mother of all RNG's" },
	{ "Type", "Multiply with carry (MWC)" },
	{ "Author", "George Marsaglia" }
} ;

#define NITEMS (sizeof(mothitems)/sizeof(mothitems[0]))

static RNGInfo mothinfo[] = { NITEMS, mothitems } ;

RNGInfo *Mother::info = mothinfo;

